const
  isUUID = require('validator/lib/isUUID'),
  isURI = require('validator/lib/isURL'),
  { DateTime } = require('luxon'),
  { ObjectUtil, uuid } = require('mbjs-utils'),
  { parseURI } = require('./lib'),
  { uuidNamespaces } = ObjectUtil,
  constants = require('./constants'),
  { JSONLD } = require('./lib/transformers')

/**
 * Default constructor
 * @param data
 */
const defaultConstructor = function (data = {}) {
  /** Check and update the supplied data */
  data.id = checkAndSetID(
    data.id && uuid.isUUID(data.id) ? undefined : data.id,
    this._path,
    checkAndSetUUID(data.id && uuid.isUUID(data.id) ? data.id : data._uuid)
  )

  if (data.created) data._created = DateTime.fromISO(data.created, {setZone: true}).toMillis()
  if (data.modified) data._modified = DateTime.fromISO(data._modified, {setZone: true}).toMillis()

  /** Populate data on the model */
  this.populate(data)
}

/**
 * All available constructors
 */
const constructors = {
  default: defaultConstructor,
  fromJSON: function (data) {
    data = JSON.parse(data)
    if (data.created) data._created = DateTime.fromISO(data.created, {setZone: true}).toMillis()
    if (data.modified) data._modified = DateTime.fromISO(data._modified, {setZone: true}).toMillis()

    this.super(data)
  }
}

/**
 * Create different UUID types
 * @param uuid
 * @param opts
 * @return {*}
 */
const checkAndSetUUID = (uuid, opts = {}) => {
  if (!uuid || !isUUID(uuid)) {
    let
      sig = opts.sig,
      ns = opts.ns || uuidNamespaces.NULL,
      { data } = opts

    if (data) sig = data.hash || data.signature || data.email
    if (sig) {
      return ObjectUtil.uuid5(sig, ns)
    }
    else if (data && data.file && data.file.hash) {
      return ObjectUtil.uuid5(data.file.hash, ns)
    }
    else {
      return ObjectUtil.uuid4()
    }
  }
  return uuid
}

/**
 * Create resource URI
 * @param id
 * @param path
 * @param uuid
 * @return {*}
 */
const checkAndSetID = (id, path, uuid) => {
  if (!id || !isURI(id)) {
    const baseURI = process.env.RESOURCE_BASE_URI || constants.BASE_URI
    id = `${baseURI}${path || ''}${uuid}`
  }
  return id
}

/**
 * Filter internal properties
 * @param obj
 * @return {*}
 */
const filterProperties = (obj) => {
  const keys = Object.keys(obj)
  for (let key of keys) {
    if (key[0] === '_') delete obj[key]
  }
  return obj
}

/**
 * All available instance methods
 */
const methods = {
  serializeJSON (opts = {}) {
    const obj = filterProperties(this.toObject())
    if (opts.pretty) return JSON.stringify(obj, null, ' ')
    return JSON.stringify(obj)
  },
  async serializeJSONLD (expanded = false) {
    if (!this['@context']) throw new Error('serializeJSONLD: No context defined')
    const obj = filterProperties(this.toObject())
    if (expanded) {
      const jsonld = new JSONLD(obj, this['@context'])
      return jsonld.toJSONLD(expanded)
    }
    return obj
  },
  async serializeRDF () {
    if (!this['@context']) throw new Error('serializeRDF: No context defined')
    const
      jsonld = new JSONLD(filterProperties(this.toObject()), this['@context']),
      doc = await jsonld.toRDF()
    return doc
  },
  touch () {
    this._updated = DateTime.local().toMillis()
  }
}

/**
 * Special fields
 */
const setSpecialFields = (resourceName, schema) => {
  return ObjectUtil.merge(schema, {
    id: {
      type: String,
      // FIXME: check why we cannot use this
      // required: true,
      default: () => ObjectUtil.uuid4(),
      transform: function (value) {
        if (uuid.isUUID(value)) {
          return `${constants.BASE_URI}${this._path}${ObjectUtil.uuid4()}`
        }
        return value
      }
    },

    /** Local document metadata **/
    _created: {
      type: Number,
      default: () => {
        return DateTime.local().toMillis()
      },
      transform: value => {
        if (typeof value === 'string') {
          return DateTime.fromISO(value).toMillis()
        }
        return value
      }
    },
    created: {
      readOnly: true,
      type: String,
      getter: function () {
        return this._created ? DateTime.fromMillis(this._created).toISO() : undefined
      }
    },
    _modified: {
      type: Number,
      transform: value => {
        if (typeof value === 'string') {
          return DateTime.fromISO(value).toMillis()
        }
        return value
      }
    },
    modified: {
      readOnly: true,
      type: String,
      getter: function () {
        return this._updated ? DateTime.fromMillis(this._updated).toISO() : undefined
      }
    },

    /** Internals */
    _uuid: {
      type: String,
      invisible: true,
      readOnly: true,
      getter: function () {
        return this.id ? parseURI(this.id).uuid : undefined
      }
    },
    _path: {
      type: String,
      invisible: true,
      readOnly: true,
      getter: () => `${resourceName}/`
    }
  })
}

module.exports = {
  methods,
  constructors,
  setSpecialFields
}
