const SchemaObject = require('schema-object')
const qs = require('qs')
const { parseValue, stringifyValue } = require('../../../lib')
const { DateTime, Duration, Interval } = require('luxon')

const Selector = new SchemaObject({
  type: {
    type: String,
    enum: ['FragmentSelector'],
    transform: value => {
      // FIXME: remove fragment selector type hack
      if (!value || value === 'Fragment') return 'FragmentSelector'
      return value
    }
  },
  value: {
    type: String,
    transform: value => {
      if (value && typeof value === 'object') {
        for (let key of Object.keys(value)) {
          if (Array.isArray(value[key])) value[key] = value[key].map(v => stringifyValue(v)).join(',')
          else value[key] = stringifyValue(value[key])
        }
        const enableEncoding = !!(process && process.env && process.env.FLAG_ENABLE_FRAGMENT_ENCODING)
        return qs.stringify(value, {
          encodeValuesOnly: enableEncoding,
          encode: enableEncoding
        })
      }
      return value
    }
  },
  conformsTo: {
    type: String,
    // transform: value => value ? value || 'http://www.w3.org/TR/media-frags/',
    enum: ['http://www.w3.org/TR/media-frags/']
  },

  /** Non-standard fields */
  _value: {
    type: 'any',
    getter: function () {
      return this.parse() || {}
    },
    readOnly: true,
    invisible: true
  },
  _valueMillis: {
    type: Number,
    readOnly: true,
    getter: function () {
      const parsed = this.parse()
      if (!parsed) return
      if (parsed['date-time:t']) {
        let parsedDate = parsed['date-time:t']
        if (Array.isArray(parsedDate) && parsedDate.length) parsedDate = parsedDate[0]
        return DateTime.fromISO(parsedDate, { setZone: true }).toMillis()
      }
      else if (parsed['t']) {
        let parsedSeconds = parsed['t']
        if (Array.isArray(parsedSeconds)) {
          parsedSeconds = parsedSeconds.map(val => parseFloat(val))
          if (parsedSeconds.length) parsedSeconds = parsedSeconds[0]
        }
        return parsedSeconds * 1000
      }
    }
  },
  _valueDuration: {
    type: Number,
    readOnly: true,
    getter: function () {
      const duration = this.getDuration()
      return duration ? duration.as('milliseconds') : undefined
    }
  }
}, {
  methods: {
    parse () {
      switch (this.type) {
        /**
         * Parse FragmentSelector
         * see https://www.w3.org/TR/annotation-model/#fragment-selector
         */
        case 'FragmentSelector':
          try {
            let { value } = this
            value = value[0] === '#' ? value.substr(1) : value
            value = value[0] === '?' ? value.substr(1) : value
            const
              enableEncoding = !!(process && process.env && process.env.FLAG_ENABLE_FRAGMENT_ENCODING),
              options = enableEncoding ? {} : { decoder: val => val },
              parsed = qs.parse(value, options),
              keys = Object.keys(parsed)

            keys.forEach(key => {
              if (typeof parsed[key] === 'string' && /,+/g.test(parsed[key])) {
                parsed[key] = parsed[key].split(',')
                  .map(v => parseValue(v))
              }
              else parsed[key] = parseValue(parsed[key])
            })

            return parsed
          }
          catch (err) {
            console.error('Selector parse error', err)
          }
          break

        /** Try to parse value if string */
        default:
          if (typeof this.value === 'string') {
            return parseValue(this.value)
          }
          return this.value
      }
    },
    getDuration () {
      if (this.type === 'FragmentSelector') {
        const parsed = this.parse()
        let value
        if (parsed['date-time:t']) {
          value = parsed['date-time:t']
          if (Array.isArray(value) && value.length === 2) {
            if (value[1] instanceof Duration) return value[1]
            if (value[0] instanceof DateTime && value[1] instanceof DateTime) {
              return Interval.fromDateTimes(value[0], value[1]).toDuration()
            }
          }
        }
        else if (parsed['t']) {
          if (Array.isArray(parsed['t'])) {
            value = Duration.fromMillis((parsed['t'][1] - parsed['t'][0]) * 1000)
          }
          else value = 0
        }
        if (value instanceof Duration) return value
      }
    }
  }
})

module.exports = Selector
