const SchemaObject = require('schema-object')
const { ObjectUtil } = require('mbjs-utils')

const constants = require('../../constants')
const { constructors, methods, setSpecialFields } = require('../../defaults')
const { SerializedObject, Stylesheet, Person } = require('../internal')
const { Target, Selector } = require('../annotation/sub-models')

const schema = setSpecialFields('maps', {
  /** Document ID & ownership **/
  '@context': {
    type: String,
    readOnly: true,
    getter: () => `${constants.BASE_URI_NS}map.jsonld`
  },
  type: [{
    type: String,
    /** Allowed types for Map */
    enum: constants.MAP_CLASSES
  }],
  creator: Person,

  /** Map title */
  title: String,
  configuration: SerializedObject,
  stylesheet: Stylesheet
})

const config = {
  strict: true,
  constructors,
  methods: ObjectUtil.merge(methods, {
    /**
     * Abstract FragmentSelector Factory
     *
     * Conforms to 'Media'
     *
     * see:
     * https://www.w3.org/TR/media-frags/
     * https://www.w3.org/TR/media-frags/#mf-advanced
     *
     * @param params
     * @param targetType - String
     */
    getTarget: function (value, targetType = constants.mapTypes.MAP_TYPE_MAP) {
      return new Target({
        type: targetType,
        id: this.id,
        selector: new Selector({
          value,
          type: 'FragmentSelector'
        })
      })
    },
    /**
     * Generate Timeline interval selector
     *
     * @param time - DateTime|Date|String
     * @param end - DateTime|Date|String
     * @return {Target}
     */
    getInterval: function (time, end = undefined) {
      const params = { 'date-time:t': time }
      if (end) {
        params['date-time:t'] = [time, end]
      }
      return this.getTarget(params, constants.mapTypes.MAP_TYPE_TIMELINE)
    },
    /**
     * Generate 2D Grid area selector
     *
     * @param point
     * @param size - Array
     * @return {Target}
     */
    get2DArea: function (point, size = undefined) {
      let params
      if (Array.isArray(point) && Array.isArray(size)) {
        params = {
          xywh: point.concat(size.slice(0, 2))
        }
      }
      else if (Array.isArray(point)) {
        params = {
          xy: point.slice(0, 2)
        }
      }
      return this.getTarget(params, constants.mapTypes.MAP_TYPE_GRID)
    },
    /**
     * Generate 3D Grid area selector
     *
     * @param point
     * @param size
     * @return {Target}
     */
    get3DArea: function (point, size = undefined) {
      let params
      if (Array.isArray(point) && Array.isArray(size)) {
        params = {
          xyzwhd: point.slice(0, 3).concat(size.slice(0, 3))
        }
      }
      else if (Array.isArray(point)) {
        params = {
          xyz: point.slice(0, 3)
        }
      }
      return this.getTarget(params, constants.mapTypes.MAP_TYPE_GRID)
    }
  })
}

const Map = new SchemaObject(schema, config)
Map.schema = schema

module.exports = Map
